@@ -1,5 +1,6 @@
 /* Open and close files for bison,
-   Copyright 1984, 1986, 1989, 1992, 2000, 2001 Free Software Foundation, Inc.
+   Copyright 1984, 1986, 1989, 1992, 2000, 2001, 2002
+   Free Software Foundation, Inc.
 
    This file is part of Bison, the GNU Compiler Compiler.
 
@@ -22,10 +23,15 @@
 #include "system.h"
 #include "getargs.h"
 #include "files.h"
-#include "xalloc.h"
 #include "gram.h"
+#include "error.h"
 #include "complain.h"
 
+/* From basename.c.  Almost a lie, as it returns a char *. */
+const char *base_name PARAMS ((char const *name));
+/* From xstrndup.c.  */
+char *xstrndup PARAMS ((const char *s, size_t n));
+
 FILE *finput = NULL;
 
 struct obstack action_obstack;
@@ -34,16 +40,18 @@ struct obstack table_obstack;
 struct obstack defines_obstack;
 struct obstack guard_obstack;
 struct obstack output_obstack;
-struct obstack graph_obstack;
 
 char *spec_outfile = NULL;	/* for -o. */
 char *spec_file_prefix = NULL;	/* for -b. */
 char *spec_name_prefix = NULL;	/* for -p. */
+char *spec_verbose_file = NULL;   /* for --verbose. */
+char *spec_graph_file = NULL;   /* for -g. */
+char *spec_defines_file = NULL; /* for --defines. */
 
 char *infile = NULL;
 char *attrsfile = NULL;
 
-static char *base_name = NULL;
+static char *full_base_name = NULL;
 static char *short_base_name = NULL;
 
 /* C source file extension (the parser source).  */
@@ -56,7 +64,7 @@ const char *header_extension = NULL;
 | Is SUFFIX ending STRING?  |
 `--------------------------*/
 
-static int
+int
 strsuffix (const char *string, const char *suffix)
 {
   size_t string_len = strlen (string);
@@ -85,32 +93,44 @@ stringappend (const char *string1, const char *string2)
 }
 
 
-/*---------------------------------------------------------------.
-|  Computes the macro name used to avoid double inclusion in the |
-|  header of the parser and store it in header_macro_name.       |
-`---------------------------------------------------------------*/
+/*-----------------------------------------------------------------.
+| Computes the macro name used to avoid double inclusion in the    |
+| header of the parser and store it in header_macro_name.  Be sure |
+| to produce valid CPP names (don't start with digit, remain       |
+| alphanumerical + underscore).                                    |
+`-----------------------------------------------------------------*/
 
 static char *
 compute_header_macro (void)
 {
-  int ite;
-  char *macro_name;
+  const char *prefix = "BISON_";
+  char *macro_name, *cp;
 
-  macro_name = XMALLOC (char,
-			strlen (base_name) +
-			strlen (header_extension) + 1);
+  if (spec_defines_file)
+    {
+      macro_name = XMALLOC (char,
+			    strlen (prefix) +
+			    strlen (spec_defines_file) + 1);
+      cp = stpcpy (macro_name, prefix);
+      cp = stpcpy (cp, spec_defines_file);
+    }
+  else
+    {
+      macro_name = XMALLOC (char,
+			    strlen (prefix) +
+			    strlen (full_base_name) +
+			    strlen (header_extension) + 1);
+      cp = stpcpy (macro_name, prefix);
+      cp = stpcpy (cp, full_base_name);
+      cp = stpcpy (cp, header_extension);
+    }
 
-  stpcpy (macro_name, base_name);
-  strcat (macro_name, header_extension);
+  for (cp = macro_name; *cp; ++cp)
+    if (islower (*cp))
+      *cp = toupper (*cp);
+    else if (!isalnum (*cp))
+      *cp = '_';
 
-  for (ite = 0; macro_name[ite]; ite++)
-    if (macro_name[ite] == '.')
-      macro_name[ite] = '_';
-    else
-      {
-	if (islower (macro_name[ite]))
-	  macro_name[ite] -= ('a' - 'A');
-      }
   return macro_name;
 }
 
@@ -194,17 +214,42 @@ skeleton_find (const char *envvar, const char *skeleton_name)
 {
   const char *res = getenv (envvar);
 
-#ifdef MSDOS
-  const char *cp;
-
-  /* File doesn't exist in current directory; try in INIT directory.  */
-  if (!res && (cp = getenv ("INIT")))
+#if defined (MSDOS) || defined (_WIN32)
+  const char *cp = getenv ("INIT");
+  if (!res)
     {
-      res = XMALLOC (char, strlen (cp) + strlen (skeleton_name) + 2);
-      sprintf (res, "%s%c%s", cp, '/', skeleton_name);
-    }
-#endif /* !MSDOS */
+      /* Skeleton file name without path */
+      const char *skel_name = strrchr (skeleton_name, '/');
+      if (!skel_name)
+        skel_name = strrchr (skeleton_name, '\\');
+      if (!skel_name)
+        skel_name = skeleton_name;
+      else
+        ++skel_name;
 
+      /* File doesn't exist in current directory; try in INIT directory.  */
+      if (cp)
+	{
+	  res = XMALLOC (char, strlen (cp) + strlen (skel_name) + 2);
+	  sprintf (res, "%s%c%s", cp, '\\', skel_name);
+	}
+      else if (access (skel_name, 4) == 0) /* Look in current dir. */
+        res = skel_name;
+      else
+	{
+	  /* Look in program locations dir. */
+	  extern char *program_name;
+	  cp = strrchr(program_name, '\\');
+	  if (!cp)
+	    return skeleton_name;
+	  else
+	    ++cp;
+	  res = XMALLOC (char, cp - program_name + strlen (skel_name) + 1);
+	  strncpy (res, program_name, cp - program_name);
+	  strcpy (res + (cp - program_name), skel_name);
+	}
+    }
+#endif /* defined (MSDOS) || defined (_WIN32) */
   if (!res)
     res = skeleton_name;
 
@@ -235,45 +280,14 @@ tr (const char *in, char from, char to)
   return (temp);
 }
 
-/* Gets the extension index in FILENAME. Returns 0 if fails to
-   find an extension.  */
-static int
-get_extension_index (const char *filename)
-{
-  int len;
-
-  len = strlen (filename);
-
-  if (filename[len-- - 1] == '.')
-    return (0);
-
-  while ((len > 0) && (filename[len - 1] != '.'))
-    if (filename[len - 1] == '/')
-      return (0);
-    else
-      len--;
-
-  return (len - 1);
-}
-
 /* Computes extensions from the grammar file extension. */
 static void
 compute_exts_from_gf (const char *ext)
 {
-  /* Checks if SRC_EXTENSION is NULL. In the other case, %source_extension
-     was specified in the grammar file.  */
-  if (src_extension == NULL)
-    {
-      src_extension = tr (ext, 'y', 'c');
-      src_extension = tr (src_extension, 'Y', 'C');
-    }
-  /* Checks if HEADER_EXTENSION is NULL. In the other case,
-     %header_extension was specified in the grammar file.  */
-  if (header_extension == NULL)
-    {
-      header_extension = tr (ext, 'y', 'h');
-      header_extension = tr (header_extension, 'Y', 'H');
-    }
+  src_extension = tr (ext, 'y', 'c');
+  src_extension = tr (src_extension, 'Y', 'C');
+  header_extension = tr (ext, 'y', 'h');
+  header_extension = tr (header_extension, 'Y', 'H');
 }
 
 /* Computes extensions from the given c source file extension. */
@@ -288,14 +302,61 @@ compute_exts_from_src (const char *ext)
   header_extension = tr (header_extension, 'C', 'H');
 }
 
+
+/* Decompose FILENAME in four parts: *BASE, *TAB, and *EXT, the fourth
+   part, (the directory) is ranging from FILENAME to the char before
+   *BASE, so we don't need an additional parameter.
+
+   *EXT points to the last period in the basename, or NULL if none.
+
+   If there is no *EXT, *TAB is NULL.  Otherwise, *TAB points to
+   `.tab' or `_tab' if present right before *EXT, or is NULL. *TAB
+   cannot be equal to *BASE.
+
+   None are allocated, they are simply pointers to parts of FILENAME.
+   Examples:
+
+   '/tmp/foo.tab.c' -> *BASE = 'foo.tab.c', *TAB = '.tab.c', *EXT =
+   '.c'
+
+   'foo.c' -> *BASE = 'foo.c', *TAB = NULL, *EXT = '.c'
+
+   'tab.c' -> *BASE = 'tab.c', *TAB = NULL, *EXT = '.c'
+
+   '.tab.c' -> *BASE = '.tab.c', *TAB = NULL, *EXT = '.c'
+
+   'foo.tab' -> *BASE = 'foo.tab', *TAB = NULL, *EXT = '.tab'
+
+   'foo_tab' -> *BASE = 'foo_tab', *TAB = NULL, *EXT = NULL
+
+   'foo' -> *BASE = 'foo', *TAB = NULL, *EXT = NULL.  */
+
+static void
+filename_split (const char *filename,
+		const char **base, const char **tab, const char **ext)
+{
+  *base = base_name (filename);
+
+  /* Look for the extension, i.e., look for the last dot. */
+  *ext = strrchr (*base, '.');
+  *tab = NULL;
+
+  /* If there is an exentension, check if there is a `.tab' part right
+     before.  */
+  if (*ext
+      && (*ext - *base) > (int) strlen (".tab")
+      && (!strncmp (*ext - strlen (".tab"), ".tab", strlen (".tab"))
+	  || !strncmp (*ext - strlen ("_tab"), "_tab", strlen ("_tab"))))
+    *tab = *ext - strlen (".tab");
+}
+
+
 /* FIXME: Should use xstrndup. */
 
 static void
 compute_base_names (void)
 {
-  size_t base_length;
-  size_t short_base_length;
-  size_t ext_index;
+  const char *base, *tab, *ext;
 
   /* If --output=foo.c was specified (SPEC_OUTFILE == foo.c),
      BASE_NAME and SHORT_BASE_NAME are `foo'.
@@ -307,92 +368,94 @@ compute_base_names (void)
      files, remove the ".c" or ".tab.c" suffix.  */
   if (spec_outfile)
     {
-#ifdef MSDOS
-      strlwr (spec_outfile);
-#endif /* MSDOS */
-      /* BASE_LENGTH includes ".tab" but not ".c".  */
-      base_length = strlen (spec_outfile);
+      filename_split (spec_outfile, &base, &tab, &ext);
 
-      ext_index = get_extension_index (spec_outfile);
-      /* If the initial segment of extension contains 'c' or a 'C', I assume
-         that it is a C or C++ source file.  */
-      if (ext_index)
-	ext_index =
-	  (strspn (spec_outfile + ext_index + 1, "cC")) ? ext_index : 0;
-      if (ext_index)
-	{
-	  base_length -= strlen (spec_outfile + ext_index);
-	  compute_exts_from_src (spec_outfile + ext_index);
-	}
+      /* The full base name goes up the EXT, excluding it. */
+      full_base_name =
+ 	xstrndup (spec_outfile,
+ 		  (strlen (spec_outfile) - (ext ? strlen (ext) : 0)));
+      /* The short base name goes up to TAB, excluding it.  */
+      short_base_name =
+ 	xstrndup (spec_outfile,
+ 		  (strlen (spec_outfile)
+ 		   - (tab ? strlen (tab) : (ext ? strlen (ext) : 0))));
 
-      base_name = strndup (spec_outfile, base_length);
-      /* SHORT_BASE_LENGTH includes neither ".tab" nor ".c".  */
-      short_base_length = base_length;
-      if (strsuffix (base_name, ".tab") || strsuffix (base_name, "_tab"))
-	short_base_length -= 4;
-      short_base_name = strndup (spec_outfile, short_base_length);
-
-      return;
+      if (ext)
+ 	compute_exts_from_src (ext);
     }
 
-  /* If --file-prefix=foo was specified, BASE_NAME and SHORT_BASE_NAME
-     are `foo'.
+  /* If --file-prefix=foo was specified, FULL_BASE_NAME = `foo.tab'
+     and SHORT_BASE_NAME = `foo'.
 
      Construct names from it.  */
-  if (spec_file_prefix)
+  else
     {
-#ifdef MSDOS
-      strlwr (spec_file_prefix);
-#endif /* MSDOS */
-      short_base_name = xstrdup (spec_file_prefix);
-      base_name = XMALLOC (char,
-			   strlen (short_base_name) + strlen (EXT_TAB) + 1);
-      stpcpy (stpcpy (base_name, short_base_name), EXT_TAB);
+      if (spec_file_prefix)
+ 	{
+ 	  /* If --file-prefix=foo was specified, SHORT_BASE_NAME =
+ 	     `foo'.  */
+ 	  short_base_name = xstrdup (spec_file_prefix);
+ 	}
+      else if (yacc_flag)
+ 	{
+ 	  /* If --yacc, then the output is `y.tab.c'. */
+ 	  short_base_name = xstrdup ("y");
+ 	}
+      else
+ 	{
+ 	  /* Otherwise, the short base name is computed from the input
+ 	     grammar: `foo.yy' => `foo'.  */
+ 	  filename_split (infile, &base, &tab, &ext);
+ 	  short_base_name =
+ 	    xstrndup (infile,
+ 		      (strlen (infile) - (ext ? strlen (ext) : 0)));
+ 	}
 
-      /* Computes the extensions from the garmmar file name.  */
-      ext_index = get_extension_index (infile);
-      /* If the initial segment of extension contains a 'y' or a 'Y', I assume
-         that it is a yacc or bison grammar file.  */
-      if (ext_index)
-	ext_index = (strspn (infile + ext_index + 1, "yY")) ? ext_index : 0;
-      if (ext_index)
-	compute_exts_from_gf (infile + ext_index);
+      /* In these cases, always append `.tab'. */
+      full_base_name = XMALLOC (char,
+ 				strlen (short_base_name)
+ 				+ strlen (EXT_TAB) + 1);
+      stpcpy (stpcpy (full_base_name, short_base_name), EXT_TAB);
 
-      return;
+      /* Computes the extensions from the grammar file name.  */
+      filename_split (infile, &base, &tab, &ext);
+      
+      if (ext && !yacc_flag)
+	compute_exts_from_gf (ext);
     }
+}
 
-  /* If neither -o nor --file-prefix were specified, and the input
-     file is foo.y, BASE_NAME is `foo.tab', and SHORT_BASE_NAME is
-     `foo'.
 
-     If --yacc is used, do as if the input file was `y.y'.  */
-  {
-    const char *name_base = yacc_flag ? "y.y" : infile;
+/*-------------------------------------------------------.
+| Close the open files, compute the output files names.  |
+`-------------------------------------------------------*/
 
-    /* BASE_LENGTH gets length of BASE_NAME, sans ".y" suffix if any.  */
+void
+compute_output_file_names (void)
+{
+  compute_base_names ();
 
-    base_length = strlen (name_base);
+  /* If not yet done. */
+  if (!src_extension)
+    src_extension = ".c";
+  if (!header_extension)
+    header_extension = ".h";
 
-    ext_index = get_extension_index (name_base);
-    /* If the initial segment of extension contains a 'y' or a 'Y', I assume
-       that it is a yacc or bison grammar file.  */
-    if (ext_index)
-      ext_index = (strspn (name_base + ext_index + 1, "yY")) ? ext_index : 0;
-    if (ext_index)
-      {
-	base_length -= strlen (name_base + ext_index);
-	compute_exts_from_gf (name_base + ext_index);
-      }
+  /* It the defines filename if not given, we create it.  */
+  if (!spec_defines_file)
+    spec_defines_file = stringappend (full_base_name, header_extension);
 
-    short_base_length = base_length;
-    short_base_name = strndup (name_base, short_base_length);
+  /* It the graph filename if not given, we create it.  */
+  if (!spec_graph_file)
+    spec_graph_file = stringappend (short_base_name, ".vcg");
 
-    base_name = XMALLOC (char,
-			 strlen (short_base_name) + strlen (EXT_TAB) + 1);
-    stpcpy (stpcpy (base_name, short_base_name), EXT_TAB);
+  spec_verbose_file = stringappend (short_base_name, EXT_OUTPUT);
+
+  attrsfile = stringappend (short_base_name, EXT_STYPE_H);
+#ifndef MSDOS
+  attrsfile = stringappend (attrsfile, header_extension);
+#endif /* MSDOS */
 
-    return;
-  }
 }
 
 /*-----------------------------------------------------------------.
@@ -412,46 +475,44 @@ open_files (void)
   obstack_init (&defines_obstack);
   obstack_init (&guard_obstack);
   obstack_init (&output_obstack);
-  obstack_init (&graph_obstack);
 }
 
 
 
-/*-----------------------------------------------------.
-| Close the open files, produce all the output files.  |
-`-----------------------------------------------------*/
+/*-----------------------.
+| Close the open file..  |
+`-----------------------*/
+
+void
+close_files (void)
+{
+  xfclose (finput);
+}
+
+/*---------------------------.
+| Produce the output files.  |
+`---------------------------*/
 
 void
 output_files (void)
 {
-  xfclose (finput);
-
-  compute_base_names ();
-
-  /* Set default extensions */
-  if (!src_extension)
-    src_extension = ".c";
-  if (!header_extension)
-    header_extension = ".h";
-
-  attrsfile = stringappend (short_base_name, EXT_STYPE_H);
-#ifndef MSDOS
-  stringappend (attrsfile, header_extension);
-#endif /* MSDOS */
-
   /* Output the main file.  */
   if (spec_outfile)
     obstack_save (&table_obstack, spec_outfile);
   else
-    obstack_save (&table_obstack, stringappend (base_name, src_extension));
+    obstack_save (&table_obstack,
+		  stringappend (full_base_name, src_extension));
+  obstack_free (&table_obstack, NULL);
 
   /* Output the header file if wanted. */
   if (defines_flag)
-    defines_obstack_save (stringappend (base_name, header_extension));
+    defines_obstack_save (spec_defines_file);
+  obstack_free (&defines_obstack, NULL);
 
   /* If we output only the table, dump the actions in ACTFILE. */
   if (no_parser_flag)
     obstack_save (&action_obstack, stringappend (short_base_name, ".act"));
+  obstack_free (&action_obstack, NULL);
 
   /* If we produced a semantic parser ATTRS_OBSTACK must be dumped
      into its own file, ATTTRSFILE.  */
@@ -460,19 +521,12 @@ output_files (void)
       char *temp_name;
 
       obstack_save (&attrs_obstack, attrsfile);
+      obstack_free (&attrs_obstack, NULL);
       temp_name = stringappend (short_base_name, EXT_GUARD_C);
 #ifndef MSDOS
       temp_name = stringappend (temp_name, src_extension);
 #endif /* MSDOS */
       obstack_save (&guard_obstack, temp_name);
+      obstack_free (&guard_obstack, NULL);
     }
-
-  if (verbose_flag)
-    /* We used to use just .out if spec_name_prefix (-p) was used, but
-       that conflicts with Posix.  */
-    obstack_save (&output_obstack,
-		  stringappend (short_base_name, EXT_OUTPUT));
-
-  if (graph_flag)
-    obstack_save (&graph_obstack, stringappend (short_base_name, ".vcg"));
 }
